const path = require('path')
const async = require('async')
const chalk = require('chalk')
const Metalsmith = require('metalsmith')
const multimatch = require('multimatch')
const render = require('consolidate').handlebars.render

const ask = require('./ask')
const logger = require('./logger')
const download = require('./download')
const getOptions = require('./options')
const localPath = require('./local-path')

/**
 * Generate a template given a `type` and `dest`.
 *
 * @param {String} type
 * @param {String} dest
 * @param {Function} done
 */

function generate(type, name, dest, done) {
    const src = localPath.getTemplatePath(type)
    const opts = getOptions(src)
    const metalsmith = Metalsmith(path.join(src, 'template'))
    const data = Object.assign(metalsmith.metadata(), {
        destDirName: name,
        inPlace: dest === process.cwd(),
        noEscape: true
    })

    const helpers = { chalk, logger }

    metalsmith.use(askQuestions(opts.prompts))
        .use(renderTemplateFiles(opts.skipInterpolation))
        .clean(false)
        .source('.') // start from template root instead of `./src` which is Metalsmith's default for `source`
        .destination(dest)
        .build((err, files) => {
            if (err) {
                done(err)
            } else {
                const hasDone = () => {
                    done()
                    if (typeof opts.complete === 'function') {
                        const helpers = { chalk, logger, files }
                        opts.complete(data, helpers)
                    } else {
                        logMessage(opts.completeMessage, data)
                    }
                }

                downloadContainer(type, metalsmith, hasDone)
            }
        })

    return data
}

/**
 * Create a middleware for asking questions.
 *
 * @param {Object} prompts
 * @return {Function}
 */

function askQuestions(prompts) {
    return (files, metalsmith, done) => {
        ask(prompts, metalsmith.metadata(), done)
    }
}

/**
 * Template in place plugin.
 *
 * @param {Object} files
 * @param {Metalsmith} metalsmith
 * @param {Function} done
 */

function renderTemplateFiles(skipInterpolation) {
    skipInterpolation = typeof skipInterpolation === 'string'
        ? [skipInterpolation]
        : skipInterpolation
    return (files, metalsmith, done) => {
        const keys = Object.keys(files)
        const metalsmithMetadata = metalsmith.metadata()
        async.each(keys, (file, next) => {
            // skipping files with skipInterpolation option
            if (skipInterpolation && multimatch([file], skipInterpolation, { dot: true }).length) {
                return next()
            }
            const str = files[file].contents.toString()
            // do not attempt to render files that do not have mustaches
            if (!/{{([^{}]+)}}/g.test(str)) {
                return next()
            }
            render(str, metalsmithMetadata, (err, res) => {
                if (err) {
                    err.message = `[${file}] ${err.message}`
                    return next(err)
                }
                files[file].contents = new Buffer(res)
                next()
            })
        }, done)
    }
}

function downloadContainer(type, metalsmith, done) {
    if (type !== 'desktop') {
        done()
    } else {
        const metadata = metalsmith.metadata()
        const name = metadata.container
        const dir = path.join(process.cwd(), metadata.destDirName, 'containers')
        download(name, dir, done)
    }
}

/**
 * Display template complete message.
 *
 * @param {String} message
 * @param {Object} data
 */

function logMessage(message, data) {
    if (!message) return
    render(message, data, (err, res) => {
        if (err) {
            console.error('\n   Error when rendering template complete message: ' + err.message.trim())
        } else {
            console.log('\n' + res.split(/\r?\n/g).map(line => '   ' + line).join('\n'))
        }
    })
}

module.exports = generate
